from .martini import Martini, GlobalProfile
from .datacube import DataCube
from ._demo import demo, demo_source
from .__version__ import __version__

__doc__ = """
    Mock APERTIF-like Radio Telescope Interferometry of the Neutral ISM

    MARTINI is a modular package for the creation of synthetic resolved HI line
    observations (data cubes) of smoothed-particle hydrodynamics simulations of
    galaxies. The various aspects of the mock-observing process are divided
    logically into sub-modules handling the data cube, source, beam, noise,
    spectral model and SPH kernel. MARTINI is object-oriented: each sub-module
    provides a class (or classes) which can be configured as desired. For most
    sub-modules, base classes are provided to allow for straightforward
    customization. Instances of each sub-module class are then given as
    parameters to the Martini class. A mock observation is then constructed by
    calling a handful of functions to execute the desired steps in the
    mock-observing process.
"""



import os
import pytest
import numpy as np
from martini.martini import Martini, GlobalProfile, _BaseMartini
from martini.datacube import DataCube, HIfreq
from martini.beams import GaussianBeam
from test_sph_kernels import simple_kernels
from martini.sph_kernels import _CubicSplineKernel, _GaussianKernel, DiracDeltaKernel
from martini.spectral_models import DiracDeltaSpectrum, GaussianSpectrum
from martini.sources import SPHSource
from astropy import units as U
from astropy.io import fits
from astropy import wcs
from astropy.coordinates import FK5, ICRS
from scipy.signal import fftconvolve


def check_mass_accuracy(m, out_mode):
    if out_mode == "hdf5":
        h5py = pytest.importorskip(
            "h5py", reason="h5py (optional dependency) not available"
        )

    # flux in channels
    F = (m.datacube._array * m.datacube.px_size**2).sum((0, 1)).squeeze()  # Jy

    # distance
    D = m.source.distance

    # channel width
    dv = np.abs(np.diff(m.datacube.velocity_channel_edges))

    # HI mass
    MHI = np.sum(
        2.36e5
        * U.Msun
        * D.to_value(U.Mpc) ** 2
        * F.to_value(U.Jy)
        * dv.to_value(U.km / U.s)
    ).to(U.Msun)

    # demand accuracy within 1% after source insertion
    assert U.isclose(MHI, m.source.mHI_g.sum(), rtol=1e-2)

    m.convolve_beam()

    # radiant intensity
    Irad = m.datacube._array.sum((0, 1)).squeeze()  # Jy / beam

    # beam area, for an equivalent Gaussian beam
    A = np.pi * m.beam.bmaj * m.beam.bmin / 4 / np.log(2) / U.beam

    # distance
    D = m.source.distance

    # channel width
    dv = np.abs(np.diff(m.datacube.velocity_channel_edges))

    # flux
    F = (Irad / A).to(U.Jy / U.arcsec**2) * m.datacube.px_size**2

    # HI mass
    MHI = np.sum(
        2.36e5
        * U.Msun
        * D.to_value(U.Mpc) ** 2
        * F.to_value(U.Jy)
        * dv.to_value(U.km / U.s)
    ).to(U.Msun)

    # demand accuracy within 1% after beam convolution
    assert U.isclose(MHI, m.source.mHI_g.sum(), rtol=1e-2)

    if out_mode == "fits":
        filename = "cube.fits"
        try:
            m.write_fits(filename)
            with fits.open(filename) as f:
                # distance
                D = m.source.distance

                # radiant intensity
                fits_wcs = wcs.WCS(f[0].header)
                Irad = U.Quantity(
                    f[0].data.T.sum((0, 1)).squeeze(), unit=f[0].header["BUNIT"]
                )

                A = (
                    np.pi
                    * (f[0].header["BMAJ"] * U.deg)
                    * (f[0].header["BMIN"] * U.deg)
                    / 4
                    / np.log(2)
                    / U.beam
                )
                px_area = U.Quantity(
                    np.abs(f[0].header["CDELT1"]), unit=f[0].header["CUNIT1"]
                ) * U.Quantity(
                    np.abs(f[0].header["CDELT2"]), unit=f[0].header["CUNIT2"]
                )

                # flux
                F = (Irad / A).to(U.Jy / U.arcsec**2) * px_area

                channel_edges = fits_wcs.sub(("spectral",)).all_pix2world(
                    np.arange(fits_wcs.sub(("spectral",)).pixel_shape[0] + 1) - 0.5,
                    0,
                ) * U.Unit(fits_wcs.wcs.cunit[fits_wcs.wcs.spec], format="fits")
                dv = np.abs(
                    np.diff(
                        channel_edges.squeeze().to(
                            U.km / U.s, equivalencies=U.doppler_radio(HIfreq)
                        )
                    )
                )

                # HI mass
                MHI = np.sum(
                    2.36e5
                    * U.Msun
                    * D.to_value(U.Mpc) ** 2
                    * F.to_value(U.Jy)
                    * dv.to_value(U.km / U.s)
                ).to(U.Msun)

            # demand accuracy within 1% in output fits file
            assert U.isclose(MHI, m.source.mHI_g.sum(), rtol=1e-2)

        finally:
            if os.path.exists(filename):
                os.remove(filename)

    if out_mode == "hdf5":
        filename = "cube.hdf5"
        try:
            m.write_hdf5(filename)
            with h5py.File(filename, "r") as f:
                # distance
                D = m.source.distance

                # radiant intensity
                Irad = U.Quantity(
                    f["FluxCube"][()].sum((0, 1)).squeeze(),
                    unit=f["FluxCube"].attrs["FluxCubeUnit"],
                )

                A = (
                    np.pi
                    * (f["FluxCube"].attrs["BeamMajor_in_deg"] * U.deg)
                    * (f["FluxCube"].attrs["BeamMinor_in_deg"] * U.deg)
                    / 4
                    / np.log(2)
                    / U.beam
                )
                dv = np.abs(
                    np.diff(
                        f["velocity_channel_edges"]
                        * U.Unit(f["velocity_channel_edges"].attrs["Unit"])
                    )
                )
                px_area = U.Quantity(
                    np.abs(f["FluxCube"].attrs["deltaRA_in_RAUnit"]),
                    unit=f["FluxCube"].attrs["RAUnit"],
                ) * U.Quantity(
                    np.abs(f["FluxCube"].attrs["deltaDec_in_DecUnit"]),
                    unit=f["FluxCube"].attrs["DecUnit"],
                )

                # flux
                F = (Irad / A).to(U.Jy / U.arcsec**2) * px_area

                # HI mass
                MHI = np.sum(
                    2.36e5
                    * U.Msun
                    * D.to_value(U.Mpc) ** 2
                    * F.to_value(U.Jy)
                    * dv.to_value(U.km / U.s)
                ).to(U.Msun)

            # demand accuracy within 1% in output hdf5 file
            assert U.isclose(MHI, m.source.mHI_g.sum(), rtol=1e-2)

        finally:
            if os.path.exists(filename):
                os.remove(filename)


class TestMartini:
    @pytest.mark.parametrize("sph_kernel", simple_kernels)
    @pytest.mark.parametrize("spectral_model", (DiracDeltaSpectrum, GaussianSpectrum))
    @pytest.mark.parametrize("out_mode", ("fits", "hdf5"))
    def test_mass_accuracy(
        self, dc_zeros, sph_kernel, spectral_model, single_particle_source, out_mode
    ):
        """
        Check that the input mass in the particles gives the correct total mass in the
        datacube, by checking the conversion back to total mass. Covers testing
        Martini.insert_source_in_cube.
        """

        hsm_g = (
            0.1 * U.kpc if sph_kernel.__name__ == "DiracDeltaKernel" else 1.0 * U.kpc
        )
        source = single_particle_source(hsm_g=hsm_g)
        # single_particle_source has a mass of 1E4Msun, temperature of 1E4K
        m = Martini(
            source=source,
            datacube=dc_zeros,
            beam=GaussianBeam(),
            noise=None,
            spectral_model=spectral_model(),
            sph_kernel=sph_kernel(),
        )
        m.insert_source_in_cube(progressbar=False)
        check_mass_accuracy(m, out_mode)

    def test_convolve_beam(self, single_particle_source):
        """
        Check that beam convolution gives result matching manual calculation.
        """
        source = single_particle_source()
        datacube = DataCube(
            n_px_x=16,
            n_px_y=16,
            n_channels=16,
            spectral_centre=source.distance * source.h * 100 * U.km / U.s / U.Mpc,
        )
        beam = GaussianBeam()
        noise = None
        sph_kernel = _GaussianKernel()
        spectral_model = GaussianSpectrum()

        m = Martini(
            source=source,
            datacube=datacube,
            beam=beam,
            noise=noise,
            sph_kernel=sph_kernel,
            spectral_model=spectral_model,
        )
        m.insert_source_in_cube()
        unconvolved_cube = m.datacube._array.copy()
        unit = unconvolved_cube.unit
        s = np.s_[..., 0] if m.datacube.stokes_axis else np.s_[...]
        for spatial_slice in iter(unconvolved_cube[s].transpose((2, 0, 1))):
            spatial_slice[...] = (
                fftconvolve(spatial_slice, m.beam.kernel, mode="same") * unit
            )
        convolved_cube = unconvolved_cube[
            m.datacube.padx : -m.datacube.padx, m.datacube.pady : -m.datacube.padx
        ]
        convolved_cube = convolved_cube.to(
            U.Jy * U.beam**-1,
            equivalencies=U.beam_angular_area(m.beam.area),
        )
        m.convolve_beam()
        assert U.allclose(m.datacube._array, convolved_cube)

    def test_add_noise(self, m_init):
        """
        Check that noise provided goes into the datacube when we call add_noise.
        """
        assert (m_init.datacube._array.sum() == 0).all()
        assert m_init.noise.seed is not None
        expected_noise = m_init.noise.generate(m_init.datacube, m_init.beam)
        m_init.noise.reset_rng()
        m_init.add_noise()
        assert U.allclose(
            m_init.datacube._array,
            expected_noise.to(
                U.Jy * U.arcsec**-2,
                equivalencies=U.beam_angular_area(m_init.beam.area),
            ).to(
                m_init.datacube._array.unit,
                equivalencies=[m_init.datacube.arcsec2_to_pix],
            ),
        )

    @pytest.mark.parametrize(
        ("ra_off", "ra_in"),
        (
            (0 * U.arcsec, True),
            (3 * U.arcsec, True),
            (9 * U.arcsec, False),
            (-3 * U.arcsec, True),
            (-9 * U.arcsec, False),
        ),
    )
    @pytest.mark.parametrize(
        ("dec_off", "dec_in"),
        (
            (0 * U.arcsec, True),
            (3 * U.arcsec, True),
            (9 * U.arcsec, False),
            (-3 * U.arcsec, True),
            (-9 * U.arcsec, False),
        ),
    )
    @pytest.mark.parametrize(
        ("v_off", "v_in"),
        (
            (0 * U.km / U.s, True),
            (3 * U.km / U.s, True),
            (7 * U.km / U.s, False),
            (-3 * U.km / U.s, True),
            (-7 * U.km / U.s, False),
        ),
    )
    @pytest.mark.parametrize(("mass_off", "mass_in"), ((0, False), (1, True)))
    @pytest.mark.parametrize("spatial", (True, False))
    @pytest.mark.parametrize("spectral", (True, False))
    @pytest.mark.parametrize("mass", (True, False))
    def test_prune_particles(
        self,
        ra_off,
        ra_in,
        dec_off,
        dec_in,
        v_off,
        v_in,
        mass_off,
        mass_in,
        single_particle_source,
        spatial,
        spectral,
        mass,
    ):
        """
        Check that a particle offset by a specific set of (RA, Dec, v) is inside/outside
        the cube as expected.
        """
        if spatial and spectral and mass:
            expect_particle = all((ra_in, dec_in, v_in, mass_in))
        elif spatial and spectral and not mass:
            expect_particle = all((ra_in, dec_in, v_in))
        elif spatial and not spectral and mass:
            expect_particle = all((ra_in, dec_in, mass_in))
        elif spatial and not spectral and not mass:
            expect_particle = all((ra_in, dec_in))
        elif not spatial and spectral and mass:
            expect_particle = all((v_in, mass_in))
        elif not spatial and spectral and not mass:
            expect_particle = v_in
        elif not spectral and not spatial and mass:
            expect_particle = mass_in
        elif not spectral and not spatial and not mass:
            expect_particle = True
        # set distance so that 1kpc = 1arcsec
        distance = (1 * U.kpc / 1 / U.arcsec).to(U.Mpc, U.dimensionless_angles())
        source = single_particle_source(
            distance=distance,
            ra=ra_off,
            dec=dec_off,
            vpeculiar=v_off,
            mHI_g=mass_off * np.ones(1) * 1.0e4 * U.Msun,
        )
        datacube = DataCube(
            n_px_x=2,
            n_px_y=2,
            n_channels=2,
            spectral_centre=source.distance * source.h * 100 * U.km / U.s / U.Mpc,
            px_size=1 * U.arcsec,
            channel_width=1 * U.km / U.s,
            ra=0 * U.deg,
            dec=0 * U.deg,
        )
        # pad size will be 5, so datacube is 12x12 pixels
        beam = GaussianBeam(bmaj=1 * U.arcsec, bmin=1 * U.arcsec, truncate=4)
        sph_kernel = _CubicSplineKernel()
        spectral_model = GaussianSpectrum(sigma=1 * U.km / U.s)
        # need to use _BaseMartini below to manipulate _prune_kwargs
        kwargs = dict(
            source=source,
            datacube=datacube,
            beam=beam,
            noise=None,
            sph_kernel=sph_kernel,
            spectral_model=spectral_model,
            _prune_kwargs=dict(spatial=spatial, spectral=spectral, mass=mass),
        )
        # if more than 1px (datacube) + 5px (pad) + 2px (sm_range) then expect to prune
        # if more than 1px (datacube) + 4px (4*spectrum_half_width) then expect to prune
        if not expect_particle:
            with pytest.raises(
                RuntimeError,
                match="No non-zero mHI source particles in target region.",
            ):
                _BaseMartini(**kwargs)
        else:
            assert _BaseMartini(**kwargs).source.npart == 1

    def test_prune_nan(self, single_particle_source):
        """
        Particles can have nan pixel coordinates if they're more than 90deg in longitude
        away from the reference pixel location of the WCS (pretty sure that's the
        condition, else maybe 90deg longitude away from the 0 pixel). We want these to
        be rejected in pruning if spatial pruning is active.

        The source has two particles, once translated to 10 Mpc distance they are at
        +10 and -10 Mpc diametrically opposed w.r.t. the observer. If we tested with just
        one invalid particle Martini exits complaining that there are no particles in the
        target region, so we put one valid and one invalid and check that only one
        remains after pruning.
        """
        source = SPHSource(
            xyz_g=U.Quantity([[0, -20], [0, 0], [0, 0]], U.Mpc),
            vxyz_g=U.Quantity([[0, 0], [0, 0], [0, 0]], U.km / U.s),
            T_g=U.Quantity([1e4, 1e4], U.K),
            hsm_g=U.Quantity([0.1, 0.1], U.kpc),
            mHI_g=U.Quantity([1e6, 1e6], U.Msun),
            ra=0 * U.deg,
            dec=0 * U.deg,
            distance=10 * U.Mpc,
        )
        datacube = DataCube(
            n_px_x=2,
            n_px_y=2,
            n_channels=2,
            channel_width=1 * U.km / U.s,
            px_size=10 * U.arcsec,
            spectral_centre=source.vsys,
            ra=source.ra + 180 * U.deg,
            dec=source.dec,
        )
        source._init_skycoords()
        source._init_pixcoords(datacube)
        print(source.pixcoords)
        assert np.isnan(source.pixcoords[0][0])
        assert np.isnan(source.pixcoords[1][0])
        m = Martini(
            source=source,
            datacube=datacube,
            beam=GaussianBeam(),
            noise=None,
            sph_kernel=DiracDeltaKernel(),
            spectral_model=DiracDeltaSpectrum(),
        )
        assert m.source.pixcoords.shape == (3, 1)

    def test_reset(self, m_nn):
        """
        Check that resetting martini instance zeros out datacube.
        """
        cube_array = m_nn.datacube._array
        assert m_nn.datacube._array.sum() > 0
        m_nn.reset()
        assert m_nn.datacube._array.sum() == 0
        # check that can start over and get the same result w/o errors
        m_nn.insert_source_in_cube(progressbar=False)
        m_nn.convolve_beam()
        assert U.allclose(cube_array, m_nn.datacube._array)
        # check that can reset after doing nothing
        m_nn.reset()
        m_nn.reset()

    def test_reset_preserves_shape(self, single_particle_source, dc_zeros):
        m = Martini(
            source=single_particle_source(),
            datacube=dc_zeros,
            beam=GaussianBeam(),
            noise=None,
            spectral_model=DiracDeltaSpectrum(),
            sph_kernel=DiracDeltaKernel(),
        )
        expected_shape = m.datacube._array.shape
        m.reset()
        assert m.datacube._array.shape == expected_shape

    def test_explicit_init_spectra(self, m_init):
        """
        Check that we can explicitly initialize the spectra (usually it happens later when
        insert_source_in_cube is called).
        """
        assert m_init.spectral_model.spectra is None
        m_init.init_spectra()
        assert m_init.spectral_model.spectra is not None

    def test_preview(self, m_init):
        """
        Simply check that the preview visualisation runs without error.
        """
        pytest.importorskip(
            "matplotlib", reason="matplotlib (optional dependency) not available."
        )
        # with default arguments
        with pytest.warns(UserWarning, match="singular"):
            # warning: single-particle source is used, so axis limits try to be equal
            m_init.preview()
        # with non-default arguments
        m_init.preview(
            max_points=1000,
            fig=2,
            lim="datacube",
            vlim="datacube",
            point_scaling="fixed",
            title="test",
        )

    def test_source_to_datacube_coord_transformation(self, single_particle_source):
        """
        Check that transformation is applied if source and datacube have different
        coordinate frames.
        """
        source = single_particle_source(hsm_g=0.01 * U.kpc)
        assert source.coordinate_frame.name == "icrs"
        datacube_icrs = DataCube(
            n_px_x=16,
            n_px_y=16,
            n_channels=16,
            channel_width=4 * U.km / U.s,
            px_size=10 * U.arcsec,
            spectral_centre=source.vsys,
            ra=source.ra,
            dec=source.dec,
            coordinate_frame=ICRS(),
        )
        m_icrs = Martini(
            source=source,
            datacube=datacube_icrs,
            beam=GaussianBeam(),
            noise=None,
            sph_kernel=DiracDeltaKernel(),
            spectral_model=DiracDeltaSpectrum(),
        )

        def centre_pixels_slice(m):
            datacube = m.datacube
            return m.datacube._array[
                datacube.n_px_x // 2
                - 1
                + datacube.padx : datacube.n_px_x // 2
                + 1
                + datacube.padx,
                datacube.n_px_y // 2
                - 1
                + datacube.pady : datacube.n_px_y // 2
                + 1
                + datacube.pady,
            ]

        assert np.sum(centre_pixels_slice(m_icrs).sum()) == 0
        m_icrs.insert_source_in_cube(progressbar=False)
        assert np.sum(centre_pixels_slice(m_icrs).sum()) > 0

        # ICRS is ~J2000 equinox. J1950 equinox is about a degree off,
        # so we should completely miss the cube (16 pix of 10 arcsec).
        datacube_fk5_J1950 = DataCube(
            n_px_x=16,
            n_px_y=16,
            n_channels=16,
            channel_width=4 * U.km / U.s,
            px_size=10 * U.arcsec,
            spectral_centre=source.vsys,
            ra=source.ra,
            dec=source.dec,
            coordinate_frame=FK5(equinox="J1950"),
        )
        with pytest.raises(
            RuntimeError,
            match="No non-zero mHI source particles in target region.",
        ):
            Martini(
                source=source,
                datacube=datacube_fk5_J1950,
                beam=GaussianBeam(),
                noise=None,
                sph_kernel=DiracDeltaKernel(),
                spectral_model=DiracDeltaSpectrum(),
            )

    def test_source_to_datacube_specsys_transformation(self, single_particle_source):
        """
        Check that spectral reference transformation is applied if source and datacube
        have different specsys.
        """
        source = single_particle_source(hsm_g=0.01 * U.kpc)
        datacube_icrs = DataCube(
            n_px_x=16,
            n_px_y=16,
            n_channels=16,
            channel_width=4 * U.km / U.s,
            px_size=10 * U.arcsec,
            spectral_centre=source.vsys,
            ra=source.ra,
            dec=source.dec,
            coordinate_frame=ICRS(),
            specsys="icrs",
        )
        m_icrs = Martini(
            source=source,
            datacube=datacube_icrs,
            beam=GaussianBeam(),
            noise=None,
            sph_kernel=DiracDeltaKernel(),
            spectral_model=DiracDeltaSpectrum(),
        )

        def centre_channels_slice(m):
            datacube = m.datacube
            return m.datacube._array[
                :, :, datacube.n_channels // 2 - 1 : datacube.n_channels // 2 + 1
            ]

        assert np.sum(centre_channels_slice(m_icrs).sum()) == 0
        m_icrs.insert_source_in_cube(progressbar=False)
        assert np.sum(centre_channels_slice(m_icrs).sum()) > 0

        # ICRS and Galactocentric are offset by many km/s depending on direction
        # so with 4 channels of 1 km/s we should completely miss the cube
        datacube_galactocentric = DataCube(
            n_px_x=16,
            n_px_y=16,
            n_channels=4,
            channel_width=1 * U.km / U.s,
            px_size=10 * U.arcsec,
            spectral_centre=source.vsys,
            ra=source.ra,
            dec=source.dec,
            coordinate_frame=ICRS(),
            specsys="galactocentric",
        )
        assert datacube_galactocentric.wcs.wcs.specsys == "galactocentric"
        with pytest.raises(
            RuntimeError,
            match="No non-zero mHI source particles in target region.",
        ):
            Martini(
                source=source,
                datacube=datacube_galactocentric,
                beam=GaussianBeam(),
                noise=None,
                sph_kernel=DiracDeltaKernel(),
                spectral_model=DiracDeltaSpectrum(),
            )

    def test_hdf5_grids(self, m):
        h5py = pytest.importorskip(
            "h5py", reason="h5py (optional dependency) not available."
        )
        origin = 0  # numpy-like, not fits-like, indexing
        filename = "cube.hdf5"
        try:
            m.write_hdf5(filename)
            with h5py.File(filename, "r") as f:
                voxel_coords = np.vstack(
                    (
                        (
                            f["RA"][()].flatten()
                            * U.Unit(f["RA"].attrs["Unit"], format="fits")
                        ).to_value(m.datacube.wcs.wcs.cunit[0]),
                        (
                            f["Dec"][()].flatten()
                            * U.Unit(f["Dec"].attrs["Unit"], format="fits")
                        ).to_value(m.datacube.wcs.wcs.cunit[1]),
                        (
                            f["channel_mids"][()].flatten()
                            * U.Unit(f["channel_mids"].attrs["Unit"], format="fits")
                        ).to_value(m.datacube.wcs.wcs.cunit[2]),
                    )
                ).T
                ra_idx, dec_idx, spec_idx = m.datacube.wcs.all_world2pix(
                    voxel_coords, origin
                ).T
                ra_idx = ra_idx.reshape((m.datacube._array.shape))
                dec_idx = dec_idx.reshape((m.datacube._array.shape))
                spec_idx = spec_idx.reshape((m.datacube._array.shape))
                expected_idx = np.meshgrid(
                    np.arange(m.datacube.n_px_x),
                    np.arange(m.datacube.n_px_y),
                    np.arange(m.datacube.n_channels),
                    indexing="ij",
                )
                assert np.allclose(ra_idx, expected_idx[0])
                assert np.allclose(dec_idx, expected_idx[1])
                assert np.allclose(spec_idx, expected_idx[2])
                vertex_coords = np.vstack(
                    (
                        (
                            f["RA_vertices"][()].flatten()
                            * U.Unit(f["RA_vertices"].attrs["Unit"], format="fits")
                        ).to_value(m.datacube.wcs.wcs.cunit[0]),
                        (
                            f["Dec_vertices"][()].flatten()
                            * U.Unit(f["Dec_vertices"].attrs["Unit"], format="fits")
                        ).to_value(m.datacube.wcs.wcs.cunit[1]),
                        (
                            f["channel_vertices"][()].flatten()
                            * U.Unit(f["channel_vertices"].attrs["Unit"], format="fits")
                        ).to_value(m.datacube.wcs.wcs.cunit[2]),
                    )
                ).T
                ra_vx_idx, dec_vx_idx, spec_vx_idx = m.datacube.wcs.all_world2pix(
                    vertex_coords, origin
                ).T
                shape = [s + 1 for s in m.datacube._array.shape]
                ra_vx_idx = ra_vx_idx.reshape(shape)
                dec_vx_idx = dec_vx_idx.reshape(shape)
                spec_vx_idx = spec_vx_idx.reshape(shape)
                expected_vx_idx = np.meshgrid(
                    np.arange(m.datacube.n_px_x + 1) - 0.5,
                    np.arange(m.datacube.n_px_y + 1) - 0.5,
                    np.arange(m.datacube.n_channels + 1) - 0.5,
                    indexing="ij",
                )
                assert np.allclose(ra_vx_idx, expected_vx_idx[0])
                assert np.allclose(dec_vx_idx, expected_vx_idx[1])
                assert np.allclose(spec_vx_idx, expected_vx_idx[2])

        finally:
            if os.path.exists(filename):
                os.remove(filename)


class TestParallel:
    def test_parallel_consistent_with_serial(self, many_particle_source, dc_zeros):
        """
        Check that running the source insertion loop in parallel gives the same result
        as running in serial.
        """
        pytest.importorskip(
            "multiprocess", reason="multiprocess (optional dependency) not available"
        )

        m = Martini(
            source=many_particle_source(),
            datacube=dc_zeros,
            beam=GaussianBeam(),
            noise=None,
            sph_kernel=_GaussianKernel(),
            spectral_model=GaussianSpectrum(),
        )

        m.insert_source_in_cube(ncpu=1, progressbar=False)
        expected_result = m.datacube._array

        # check that we're not testing on a zero array
        assert m.datacube._array.sum() > 0

        m.reset()

        # check the reset was successful
        assert np.allclose(
            m.datacube._array.to_value(m.datacube._array.unit),
            0.0,
        )

        m.insert_source_in_cube(ncpu=2, progressbar=False)

        assert U.allclose(m.datacube._array, expected_result)


class TestGlobalProfile:
    @pytest.mark.parametrize("spectral_model", (DiracDeltaSpectrum, GaussianSpectrum))
    @pytest.mark.parametrize("ra", (0 * U.deg, 180 * U.deg))
    def test_mass_accuracy(self, spectral_model, single_particle_source, ra):
        """
        Check that the input mass in the particles gives the correct total mass in the
        spectrum, by checking the conversion back to total mass. Covers testing
        GlobalProfile.insert_source_in_spectrum.
        """

        # single_particle_source has a mass of 1E4Msun, temperature of 1E4K
        # we test both ra=0deg and ra=180deg to make sure all particles included
        source = single_particle_source(ra=ra)
        m = GlobalProfile(
            source=source,
            spectral_model=spectral_model(),
            n_channels=32,
            channel_width=10 * U.km * U.s**-1,
            spectral_centre=source.vsys,
        )

        m.insert_source_in_spectrum()

        # flux
        F = m.spectrum.sum()  # Jy

        # distance
        D = m.source.distance

        # channel width
        dv = m.channel_width

        # HI mass
        MHI = (
            2.36e5
            * U.Msun
            * D.to_value(U.Mpc) ** 2
            * F.to_value(U.Jy)
            * dv.to_value(U.km / U.s)
        ).to(U.Msun)

        # demand accuracy within 1% after source insertion
        assert U.isclose(MHI, m.source.mHI_g.sum(), rtol=1e-2)

    @pytest.mark.parametrize(
        ("ra_off", "ra_in"),
        (
            (0 * U.arcsec, True),
            (3 * U.arcsec, True),
            (5 * U.deg, False),  # global profile uses 1 deg pixel
            (-3 * U.arcsec, True),
            (-5 * U.deg, False),  # global profile uses 1 deg pixel
        ),
    )
    @pytest.mark.parametrize(
        ("dec_off", "dec_in"),
        (
            (0 * U.arcsec, True),
            (3 * U.arcsec, True),
            (5 * U.deg, False),  # global profile uses 1 deg pixel
            (-3 * U.arcsec, True),
            (-5 * U.deg, False),  # global profile uses 1 deg pixel
        ),
    )
    @pytest.mark.parametrize(
        ("v_off", "v_in"),
        (
            (0 * U.km / U.s, True),
            (3 * U.km / U.s, True),
            (7 * U.km / U.s, False),
            (-3 * U.km / U.s, True),
            (-7 * U.km / U.s, False),
        ),
    )
    def test_prune_particles(
        self, ra_off, ra_in, dec_off, dec_in, v_off, v_in, single_particle_source
    ):
        """
        Check that a particle offset by a specific set of (RA, Dec, v) is inside/outside
        the cube as expected. GlobalProfile should ignore RA, Dec when pruning.
        """
        # GlobalProfile should ignore RA, Dec when pruning:
        expect_particle = v_in
        # set distance so that 1kpc = 1arcsec
        distance = (1 * U.kpc / 1 / U.arcsec).to(U.Mpc, U.dimensionless_angles())
        source = single_particle_source(
            distance=distance, ra=ra_off, dec=dec_off, vpeculiar=v_off
        )
        spectral_model = GaussianSpectrum(sigma=1 * U.km / U.s)
        kwargs = dict(
            source=source,
            spectral_model=spectral_model,
            n_channels=2,
            channel_width=1 * U.km / U.s,
            spectral_centre=source.distance * source.h * 100 * U.km / U.s / U.Mpc,
        )
        # if more than 1px (datacube) + 4px (4*spectrum_half_width) then expect to prune
        if not expect_particle:
            with pytest.raises(
                RuntimeError,
                match="No non-zero mHI source particles in target region.",
            ):
                GlobalProfile(**kwargs)
        else:
            assert GlobalProfile(**kwargs).source.npart == 1

    def test_reset(self, gp):
        """
        Check that resetting global profile instance zeros out datacube and spectrum.
        """
        cube_array = gp._datacube._array
        assert gp._datacube._array.sum() > 0
        spectrum = gp.spectrum
        assert spectrum.sum() > 0
        gp.reset()
        assert gp._datacube._array.sum() == 0
        assert not hasattr(gp, "_spectrum")
        # check that can start over and get the same result w/o errors
        gp.insert_source_in_spectrum()
        assert U.allclose(cube_array, gp._datacube._array)
        assert U.allclose(spectrum, gp.spectrum)
        # check that can reset after doing nothing
        gp.reset()
        gp.reset()

    def test_preview(self, gp):
        """
        Simply check that the preview visualisation runs without error.
        """
        pytest.importorskip(
            "matplotlib", reason="matplotlib (optional dependency) not available."
        )
        # with default arguments
        with pytest.warns(UserWarning, match="singular"):
            # warning: single-particle source is used, so axis limits try to be equal
            gp.preview()
        # with non-default arguments
        gp.preview(
            max_points=1000,
            fig=2,
            lim="datacube",
            vlim="datacube",
            point_scaling="fixed",
            title="test",
        )

    def test_channel_modes(self, single_particle_source):
        """
        Check that channels have expected units in both modes (frequency, velocity).
        """
        source = single_particle_source()
        channel_width = 10 * U.km * U.s**-1
        m = GlobalProfile(
            source=source,
            spectral_model=GaussianSpectrum(sigma="thermal"),
            n_channels=32,
            channel_width=channel_width,
            spectral_centre=source.vsys,
        )
        expected_units = channel_width.unit
        # these will raise if there's a problem:
        m.channel_edges.to(expected_units)
        m.channel_mids.to(expected_units)

    def test_view_spectrum(self, gp):
        """
        Simply check that plotting spectrum runs without error.
        """
        pytest.importorskip(
            "matplotlib", reason="matplotlib (optional dependency) not available."
        )
        # with default arguments
        gp.plot_spectrum()
        # with non-default arguments
        gp.plot_spectrum(fig=2, title="test", show_vsys=False)


class TestMartiniWithDataCubeFromWCS:
    @pytest.mark.parametrize("out_mode", ("fits", "hdf5"))
    def test_source_insertion(self, dc_wcs, single_particle_source, out_mode):
        datacube = dc_wcs
        distance = (
            datacube.spectral_centre.to(
                U.km / U.s, equivalencies=U.doppler_radio(HIfreq)
            )
            / (70 * U.km / U.s / U.Mpc)
        ).to(U.Mpc)
        source = single_particle_source(
            ra=datacube.ra,
            dec=datacube.dec,
            distance=distance,
            hsm_g=(3 * datacube.px_size * distance).to(
                U.kpc, equivalencies=U.dimensionless_angles()
            ),
        )
        beam = GaussianBeam(bmaj=3 * datacube.px_size, bmin=3 * datacube.px_size)
        m = Martini(
            source=source,
            datacube=datacube,
            beam=beam,
            noise=None,
            spectral_model=GaussianSpectrum(sigma="thermal"),
            sph_kernel=_CubicSplineKernel(),
        )
        m.insert_source_in_cube(progressbar=False)
        check_mass_accuracy(m, out_mode)



import typing as T
import astropy.units as U
from astropy.wcs.wcs import WCS
from astropy.coordinates.builtin_frames.baseradec import BaseRADecFrame

HIfreq: U.Quantity[U.Hz]

class DataCube:
    px_size: U.Quantity[U.arcsec]
    arcsec2_to_pix: T.Tuple[
        U.Quantity[U.Jy * U.pix**-2],
        U.Quantity[U.Jy * U.arcsec**-2],
        T.Callable[[U.Quantity[U.Jy * U.pix**-2]], U.Quantity[U.Jy * U.arcsec**-2]],
        T.Callable[[U.Quantity[U.Jy * U.arcsec**-2]], U.Quantity[U.Jy * U.pix**-2]],
    ]
    channel_width: U.Quantity[U.km / U.s]
    spectral_centre: U.Quantity[U.km / U.s]
    ra: U.Quantity[U.deg]
    dec: U.Quantity[U.deg]
    padx: int
    pady: int
    _array: (
        U.Quantity[U.Jy * U.pix**-2]
        | U.Quantity[U.Jy * U.arcsec**-2]
        | U.Quantity[U.Jy * U.beam**-1]
    )
    _wcs: T.Optional[WCS]
    n_px_x: int
    n_px_y: int
    n_channels: int
    stokes_axis: bool
    coordinate_frame: BaseRADecFrame
    specsys: str
    _freq_channel_mode: bool
    _channel_edges: T.Optional[T.Union[U.Quantity[U.Hz], U.Quantity[U.m / U.s]]]
    _channel_mids: T.Optional[T.Union[U.Quantity[U.Hz], U.Quantity[U.m / U.s]]]

    def __init__(
        self,
        n_px_x: int = ...,
        n_px_y: int = ...,
        n_channels: int = ...,
        px_size: U.Quantity[U.arcsec] = ...,
        channel_width: U.Quantity[U.arcsec] = ...,
        spectral_centre: U.Quantity[U.km / U.s] = ...,
        ra: U.Quantity[U.deg] = ...,
        dec: U.Quantity[U.deg] = ...,
        stokes_axis: bool = ...,
        coordinate_frame: BaseRADecFrame = ...,
        specsys: str = ...,
        velocity_centre: None = ...,  # deprecated
    ) -> None: ...
    @classmethod
    def from_wcs(cls, input_wcs: WCS, specsys=T.Optional[str]) -> T.Self: ...
    @property
    def units(
        self,
    ) -> T.Union[
        T.Tuple[
            U.Quantity[U.deg],
            U.Quantity[U.deg],
            T.Union[U.Quantity[U.Hz], U.Quantity[U.m / U.s]],
        ],
        T.Tuple[
            U.Quantity[U.deg],
            U.Quantity[U.deg],
            T.Union[U.Quantity[U.Hz], U.Quantity[U.m / U.s]],
            U.Quantity[U.dimensionless_unscaled],
        ],
    ]: ...
    def velocity_channels(self) -> None: ...  # deprecated
    def freq_channels(self) -> None: ...  # deprecated
    @property
    def wcs(self) -> WCS: ...
    @property
    def channel_mids(self) -> T.Union[U.Quantity[U.Hz], U.Quantity[U.m / U.s]]: ...
    @property
    def channel_edges(self) -> T.Union[U.Quantity[U.Hz], U.Quantity[U.m / U.s]]: ...
    @property
    def velocity_channel_mids(self) -> U.Quantity[U.m / U.s]: ...
    @property
    def velocity_channel_edges(self) -> U.Quantity[U.m / U.s]: ...
    @property
    def frequency_channel_mids(self) -> U.Quantity[U.Hz]: ...
    @property
    def frequency_channel_edges(self) -> U.Quantity[U.Hz]: ...
    @property
    def _stokes_index(self) -> T.Optional[int]: ...
    @property
    def channel_maps(self) -> T.Iterator[U.Quantity]: ...
    @property
    def spatial_slices(self) -> T.Iterator[U.Quantity]: ...
    @property
    def spectra(self) -> T.Iterator[U.Quantity]: ...
    def add_pad(self, pad: T.Tuple[int, int]) -> None: ...
    def drop_pad(self) -> None: ...
    def copy(self) -> T.Self: ...
    def save_state(self, filename: str, overwrite: bool = ...) -> None: ...
    @classmethod
    def load_state(cls, filename: str) -> T.Self: ...
    def __repr__(self) -> str: ...

class _GlobalProfileDataCube(DataCube):
    def __init__(
        self,
        n_channels: int = ...,
        channel_width: U.Quantity[U.arcsec] = ...,
        spectral_centre: U.Quantity[U.km / U.s] = ...,
        specsys: str = ...,
        velocity_centre: None = ...,  # deprecated
    ) -> None: ...



import abc
import numpy as np
from _typeshed import Incomplete
from abc import abstractmethod
from martini.datacube import DataCube as DataCube
from martini.sources.sph_source import SPHSource as SPHSource
import typing as T
import astropy.units as U
from types import EllipsisType

class _BaseSpectrum(metaclass=abc.ABCMeta):
    __metaclass__: Incomplete
    spectra: T.Optional[U.Quantity[U.Jy]]
    ncpu: int
    spec_dtype: type
    spectral_function_extra_data: T.Optional[T.Dict[str, T.Any]]

    def __init__(self, ncpu: T.Optional[int] = ..., spec_dtype: type = ...) -> None: ...
    def init_spectra(self, source: SPHSource, datacube: DataCube) -> None: ...
    def evaluate_spectra(
        self,
        source: SPHSource,
        datacube: DataCube,
        mask: T.Union[slice, EllipsisType] = ...,
    ) -> None: ...
    @abstractmethod
    def half_width(self, source: SPHSource) -> U.Quantity[U.km / U.s]: ...
    @abstractmethod
    def spectral_function(
        self,
        a: U.Quantity[U.km / U.s],
        b: U.Quantity[U.km / U.s],
        vmids: U.Quantity[U.km / U.s],
    ) -> U.Quantity[U.dimensionless_unscaled]: ...
    def init_spectral_function_extra_data(
        self,
        source: SPHSource,
        datacube: DataCube,
        mask: T.Union[slice, EllipsisType] = ...,
    ) -> None: ...

class GaussianSpectrum(_BaseSpectrum):
    sigma_mode: T.Union[str, U.Quantity[U.km / U.s]]

    def __init__(
        self,
        sigma: T.Union[str, U.Quantity[U.km / U.s]] = ...,
        ncpu: T.Optional[int] = None,
        spec_dtype: type = ...,
    ) -> None: ...
    def spectral_function(
        self,
        a: U.Quantity[U.km / U.s],
        b: U.Quantity[U.km / U.s],
        vmids: U.Quantity[U.km / U.s],
    ) -> U.Quantity[U.dimensionless_unscaled]: ...
    def init_spectral_function_extra_data(
        self,
        source: SPHSource,
        datacube: DataCube,
        mask: T.Union[slice, EllipsisType] = ...,
    ) -> None: ...
    def half_width(self, source: SPHSource) -> U.Quantity[U.km / U.s]: ...

class DiracDeltaSpectrum(_BaseSpectrum):
    def __init__(self, ncpu: T.Optional[int] = ..., spec_dtype: type = ...) -> None: ...
    def spectral_function(
        self,
        a: U.Quantity[U.km / U.s],
        b: U.Quantity[U.km / U.s],
        vmids: U.Quantity[U.km / U.s],
    ) -> U.Quantity[U.dimensionless_unscaled]: ...
    def half_width(self, source: SPHSource) -> U.Quantity[U.km / U.s]: ...



import pytest
import numpy as np
from martini import DataCube
from martini.spectral_models import GaussianSpectrum, DiracDeltaSpectrum
from astropy import units as U

spectral_models = GaussianSpectrum, DiracDeltaSpectrum


class TestGaussianSpectrum:
    @pytest.mark.parametrize("sigma", ("thermal", 7.0 * U.km / U.s))
    def test_init_spectra(self, sigma, single_particle_source):
        """
        Check that spectrum sums to expected flux.
        """
        source = single_particle_source(distance=1 * U.Mpc)  # D=1Mpc
        source._init_skycoords()
        spectral_model = GaussianSpectrum(sigma=sigma)
        datacube = DataCube(
            n_channels=64, channel_width=4 * U.km / U.s, spectral_centre=source.vsys
        )
        spectral_model.init_spectra(source, datacube)
        expected_flux = (
            source.mHI_g[0] / 2.36e5 * U.Jy * U.km * U.s**-1 / U.Msun
        )  # D=1Mpc
        flux = spectral_model.spectra[0].sum() * datacube.channel_width
        assert U.isclose(flux, expected_flux, rtol=1.0e-5)

    def test_half_width_constant(self, single_particle_source):
        """
        Check that spectrum reports expected width (constant case).
        """
        sigma = 7.0 * U.km / U.s
        source = single_particle_source()
        spectral_model = GaussianSpectrum(sigma=sigma)
        assert U.isclose(spectral_model.half_width(source), sigma)

    def test_half_width_thermal(self, single_particle_source):
        """
        Check that spectrum reports expected half width (thermal case).
        """
        expected_sigma = 9.0853727258 * U.km / U.s  # @1E4K
        source = single_particle_source()
        spectral_model = GaussianSpectrum(sigma="thermal")
        assert U.isclose(spectral_model.half_width(source)[0], expected_sigma)

    @pytest.mark.parametrize("sigma", ("thermal", 7.0 * U.km / U.s))
    def test_spectral_function(self, sigma, single_particle_source):
        """
        Check that spectral function gives a normalised spectrum.
        """
        source = single_particle_source()
        source._init_skycoords()
        spectral_model = GaussianSpectrum(sigma=sigma)
        datacube = DataCube(
            n_channels=64, channel_width=4 * U.km / U.s, spectral_centre=source.vsys
        )
        spectral_model.init_spectral_function_extra_data(source, datacube)
        spectrum = spectral_model.spectral_function(
            datacube.channel_edges[np.newaxis, 1:],
            datacube.channel_edges[np.newaxis, :-1],
            # expected from single_particle_source:
            U.Quantity([source.vsys])[:, np.newaxis],
        )
        assert U.isclose(spectrum.sum(), 1.0 * U.dimensionless_unscaled, rtol=1.0e-4)

    @pytest.mark.parametrize("sigma", ("thermal", 7.0 * U.km / U.s))
    @pytest.mark.parametrize("source", ("single_particle_source", "cross_source"))
    def test_spectral_function_extra_data(self, sigma, source, request):
        """
        Check that required extra data is loaded: velocity dispersions.
        """
        source = request.getfixturevalue(source)()
        source._init_skycoords()
        spectral_model = GaussianSpectrum(sigma=sigma)
        datacube = DataCube(
            n_channels=64, channel_width=4 * U.km / U.s, spectral_centre=source.vsys
        )
        spectral_model.init_spectral_function_extra_data(source, datacube)
        extra_data = spectral_model.spectral_function_extra_data
        assert set(extra_data.keys()) == {"sigma"}
        assert U.allclose(
            extra_data["sigma"].squeeze(), spectral_model.half_width(source)
        )
        expected_shape = tuple() if sigma != "thermal" else (source.npart, 1)
        assert extra_data["sigma"].shape == expected_shape


class TestDiracDeltaSpectrum:
    def test_init_spectra(self, single_particle_source):
        """
        Chec that spectrum sums to expected flux.
        """
        source = single_particle_source(distance=1 * U.Mpc)  # D=1Mpc
        source._init_skycoords()
        spectral_model = DiracDeltaSpectrum()
        datacube = DataCube(
            n_channels=64, channel_width=4 * U.km / U.s, spectral_centre=source.vsys
        )
        spectral_model.init_spectra(source, datacube)
        expected_flux = (
            source.mHI_g[0] / 2.36e5 * U.Jy * U.km * U.s**-1 / U.Msun
        )  # D=1Mpc
        flux = spectral_model.spectra[0].sum() * datacube.channel_width
        assert U.isclose(flux, expected_flux, rtol=1.0e-5)

    def test_half_width(self, single_particle_source):
        """
        Check that spectrum reports zero width.
        """
        source = single_particle_source()
        spectral_model = DiracDeltaSpectrum()
        assert U.isclose(spectral_model.half_width(source), 0 * U.km / U.s)

    def test_spectral_function(self, single_particle_source):
        """
        Check that spectral function returns normalised spectrum.
        """
        source = single_particle_source()
        source._init_skycoords()
        spectral_model = DiracDeltaSpectrum()
        datacube = DataCube(
            n_channels=64, channel_width=4 * U.km / U.s, spectral_centre=source.vsys
        )
        spectrum = spectral_model.spectral_function(
            datacube.velocity_channel_edges[1:],
            datacube.velocity_channel_edges[:-1],
            U.Quantity([source.vsys]),  # expected from _SingleParticleSource
        )
        assert U.isclose(spectrum.sum(), 1.0 * U.dimensionless_unscaled, rtol=1.0e-4)

    def test_spectral_function_extra_data(self, single_particle_source):
        """
        Check that no extra data is loaded.
        """
        source = single_particle_source()
        source._init_skycoords()
        datacube = DataCube(
            n_channels=64, channel_width=4 * U.km / U.s, spectral_centre=source.vsys
        )
        spectral_model = DiracDeltaSpectrum()
        spectral_model.init_spectral_function_extra_data(source, datacube)
        extra_data = spectral_model.spectral_function_extra_data
        assert len(extra_data) == 0


class TestSpectrumPrecision:
    @pytest.mark.parametrize("SpectralModel", spectral_models)
    @pytest.mark.parametrize("dtype", (np.float64, np.float32))
    def test_spectrum_precision(self, SpectralModel, dtype, single_particle_source):
        """
        Check that spectral model can use specified precision.
        """
        source = single_particle_source()
        source._init_skycoords()
        spectral_model = SpectralModel(spec_dtype=dtype)
        datacube = DataCube()
        spectral_model.init_spectra(source, datacube)
        assert spectral_model.spectra.dtype == dtype


class TestParallelSpectra:
    @pytest.mark.parametrize("SpectralModel", spectral_models)
    def test_parallel_spectra(self, SpectralModel, cross_source):
        """
        Check that spectra calculated in serial and parallel are consistent.
        """
        pytest.importorskip(
            "multiprocess", reason="multiprocess (optional dependency) not available."
        )
        source = cross_source()
        source._init_skycoords()
        spectral_model_serial = SpectralModel()
        spectral_model_parallel = SpectralModel(ncpu=2)
        datacube = DataCube()
        spectral_model_serial.init_spectra(source, datacube)
        spectral_model_parallel.init_spectra(source, datacube)
        assert U.allclose(
            spectral_model_serial.spectra, spectral_model_parallel.spectra
        )



"""
Provides the generic :class:`~martini.sources.sph_source.SPHSource` class for working
with any SPH or similar simulation as input.
"""

import numpy as np
from astropy.coordinates import (
    CartesianRepresentation,
    CartesianDifferential,
    SphericalRepresentation,
    SkyCoord,
    SpectralCoord,
    ICRS,
)
from astropy.coordinates.matrix_utilities import rotation_matrix
import astropy.units as U
from ._L_align import L_align
from ._cartesian_translation import translate, translate_d
from ..datacube import HIfreq

# Extend CartesianRepresentation to allow coordinate translation
setattr(CartesianRepresentation, "translate", translate)

# Extend CartesianDifferential to allow velocity (or other differential)
# translation
setattr(CartesianDifferential, "translate", translate_d)

_origin = CartesianRepresentation(
    np.zeros((3, 1)) * U.kpc,
    differentials={"s": CartesianDifferential(np.zeros((3, 1)) * U.km / U.s)},
)


class SPHSource(object):
    """
    Class abstracting HI emission sources consisting of SPH simulation
    particles.

    This class constructs an HI emission source from arrays of SPH particle
    properties: mass, smoothing length, temperature, position, and velocity.

    Parameters
    ----------
    distance : ~astropy.units.Quantity, optional
        :class:`~astropy.units.Quantity`, with dimensions of length.
        Source distance, also used to set the velocity offset via Hubble's law.
        (Default: ``3 * U.Mpc``)

    vpeculiar : ~astropy.units.Quantity, optional
        :class:`~astropy.units.Quantity`, with dimensions of velocity.
        Source peculiar velocity along the direction to the source centre.
        (Default: ``0 * U.km * U.s**-1``)

    rotation : dict, optional
        Must have a single key, which must be one of ``axis_angle``, ``rotmat`` or
        ``L_coords``. Note that the 'y-z' plane will be the one eventually placed in the
        plane of the "sky". The corresponding value must be:

        - ``axis_angle`` : 2-tuple, first element one of 'x', 'y', 'z' for the \
        axis to rotate about, second element a :class:`~astropy.units.Quantity` with \
        dimensions of angle, indicating the angle to rotate through.
        - ``rotmat`` : A (3, 3) :class:`~numpy.ndarray` specifying a rotation.
        - ``L_coords`` : A 2-tuple containing an inclination and an azimuthal \
        angle (both :class:`~astropy.units.Quantity` instances with dimensions of \
        angle). The routine will first attempt to identify a preferred plane \
        based on the angular momenta of the central 1/3 of particles in the \
        source. This plane will then be rotated to lie in the plane of the \
        "sky" ('y-z'), rotated by the azimuthal angle about its angular \
        momentum pole (rotation about 'x'), and inclined (rotation about \
        'y'). A 3-tuple may be provided instead, in which case the third \
        value specifies the position angle on the sky (second rotation about 'x'). \
        The default position angle is 270 degrees.

        (Default: ``np.eye(3)``)

    ra : ~astropy.units.Quantity, optional
        :class:`~astropy.units.Quantity`, with dimensions of angle.
        Right ascension for the source centroid. (Default: ``0 * U.deg``)

    dec : ~astropy.units.Quantity, optional
        :class:`~astropy.units.Quantity`, with dimensions of angle.
        Declination for the source centroid. (Default: ``0 * U.deg``)

    h : float, optional
        Dimensionless hubble constant,
        :math:`H_0 = h (100\\,\\mathrm{km}\\,\\mathrm{s}^{-1}\\,\\mathrm{Mpc}^{-1})`.
        (Default: ``0.7``)

    T_g : ~astropy.units.Quantity
        :class:`~astropy.units.Quantity`, with dimensions of temperature.
        Particle temperature.

    mHI_g : ~astropy.units.Quantity
        :class:`~astropy.units.Quantity`, with dimensions of mass.
        Particle HI mass.

    xyz_g : ~astropy.units.Quantity
        :class:`~astropy.units.Quantity` , with dimensions of length.
        Particle position offset from source centroid. Note that the 'y-z'
        plane is that eventually placed in the plane of the "sky"; 'x' is
        the axis corresponding to the "line of sight".

    vxyz_g : ~astropy.units.Quantity
        :class:`~astropy.units.Quantity`, with dimensions of velocity.
        Particle velocity offset from source centroid. Note that the 'y-z'
        plane is that eventually placed in the plane of the "sky"; 'x' is
        the axis corresponding to the "line of sight".

    hsm_g : ~astropy.units.Quantity
        :class:`~astropy.units.Quantity`, with dimensions of length.
        Particle SPH smoothing lengths, defined as the FWHM of the smoothing kernel.
        Smoothing lengths are variously defined in the literature as the radius where
        the kernel amplitude reaches 0, or some rational fraction of this radius (and
        other definitions may well exist). The FWHM requested here is not a standard
        choice (with the exception of SWIFT snapshots!), but has the advantage of avoiding
        ambiguity in the definition.

    coordinate_axis : int, optional
        Rank of axis corresponding to position or velocity of a single
        particle. I.e. ``coordinate_axis=0`` if shape is (3, N), or ``1`` if (N, 3).
        Usually prefer to omit this as it can be determined automatically, but is
        ambiguous for sources with exactly 3 particles. (Default: ``None``)

    coordinate_frame : ~astropy.coordinates.builtin_frames.baseradec.BaseRADecFrame, \
    optional
        The coordinate frame assumed in converting particle coordinates to RA and Dec, and
        for transforming coordinates and velocities to the data cube frame. The frame
        needs to have a well-defined velocity as well as spatial origin. Recommended
        frames are :class:`~astropy.coordinates.GCRS`, :class:`~astropy.coordinates.ICRS`,
        :class:`~astropy.coordinates.HCRS`, :class:`~astropy.coordinates.LSRK`,
        :class:`~astropy.coordinates.LSRD` or :class:`~astropy.coordinates.LSR`. The frame
        should be passed initialized, e.g. ``ICRS()`` (not just ``ICRS``).
        (Default: ``astropy.coordinates.ICRS()``)
    """

    def __init__(
        self,
        distance=3.0 * U.Mpc,
        vpeculiar=0.0 * U.km / U.s,
        rotation={"rotmat": np.eye(3)},
        ra=0.0 * U.deg,
        dec=0.0 * U.deg,
        h=0.7,
        T_g=None,
        mHI_g=None,
        xyz_g=None,
        vxyz_g=None,
        hsm_g=None,
        coordinate_axis=None,
        coordinate_frame=ICRS(),
    ):
        if coordinate_axis is None:
            if (xyz_g.shape[0] == 3) and (xyz_g.shape[1] != 3):
                coordinate_axis = 0
            elif (xyz_g.shape[0] != 3) and (xyz_g.shape[1] == 3):
                coordinate_axis = 1
            elif xyz_g.shape == (3, 3):
                raise RuntimeError(
                    "martini.sources.SPHSource: cannot guess "
                    "coordinate_axis with shape (3, 3), provide"
                    " explicitly."
                )
            else:
                raise RuntimeError(
                    "martini.sources.SPHSource: incorrect "
                    "coordinate shape (not (3, N) or (N, 3))."
                )
        else:
            coordinate_axis = coordinate_axis

        if xyz_g.shape != vxyz_g.shape:
            raise ValueError(
                "martini.sources.SPHSource: xyz_g and vxyz_g must"
                " have matching shapes."
            )

        if coordinate_axis == 0:
            xyz_g = xyz_g.T
            vxyz_g = vxyz_g.T
        self.coordinate_frame = coordinate_frame
        self.h = h
        self.T_g = T_g
        self.mHI_g = mHI_g
        self.input_mass = mHI_g.sum()
        self.coordinates_g = CartesianRepresentation(
            xyz_g,
            xyz_axis=1,
            differentials={"s": CartesianDifferential(vxyz_g, xyz_axis=1)},
        )
        self.hsm_g = hsm_g

        self.npart = self.coordinates_g.size

        self.ra = ra
        self.dec = dec
        self.distance = distance
        self.vpeculiar = vpeculiar
        self.vhubble = (self.h * 100.0 * U.km * U.s**-1 * U.Mpc**-1) * self.distance
        self.vsys = self.vhubble + self.vpeculiar
        self.current_rotation = np.eye(3)
        self.rotate(**rotation)
        self.skycoords = None
        self.spectralcoords = None
        self.pixcoords = None
        return

    def _init_skycoords(self, _reset=True):
        """
        Initialize the sky coordinates of the particles.

        Parameters
        ----------
        _reset : bool
            If ``True``, return particles to their original positions. Setting to
            ``False`` is only intended for testing. (Default: ``True``)
        """
        # _reset False only for unit testing
        distance_unit_vector = (
            SphericalRepresentation(self.ra, self.dec, 1)
            .represent_as(CartesianRepresentation)
            .xyz
        )
        distance_vector = distance_unit_vector * self.distance
        vpeculiar_vector = distance_unit_vector * self.vpeculiar

        self.rotate(axis_angle=("y", -self.dec))
        self.rotate(axis_angle=("z", self.ra))
        self.translate(distance_vector)
        # must be after translate:
        # \vec{v_H} = (100 h km/s/Mpc * D) * r^, but D * r^ is just \vec{r}:
        vhubble_vectors = (
            self.h * 100.0 * U.km * U.s**-1 * U.Mpc**-1
        ) * self.coordinates_g.xyz
        self.boost(vpeculiar_vector)
        # can't use boost for particle-by-particle velocities:
        self.coordinates_g.differentials["s"] = CartesianDifferential(
            self.coordinates_g.differentials["s"].d_xyz + vhubble_vectors
        )
        self.skycoords = SkyCoord(
            self.coordinates_g, frame=self.coordinate_frame, copy=True
        )
        origin_skycoord = SkyCoord(
            x=0 * U.kpc,
            y=0 * U.kpc,
            z=0 * U.kpc,
            v_x=0 * U.km / U.s,
            v_y=0 * U.km / U.s,
            v_z=0 * U.km / U.s,
            representation_type="cartesian",
            differential_type="cartesian",
            frame=self.coordinate_frame,
        )
        self.spectralcoords = SpectralCoord(
            self.skycoords.radial_velocity,
            doppler_convention="radio",
            doppler_rest=HIfreq,
            target=self.skycoords,
            observer=origin_skycoord,
        )
        if _reset:
            self.coordinates_g.differentials["s"] = CartesianDifferential(
                self.coordinates_g.differentials["s"].d_xyz - vhubble_vectors
            )
            self.boost(-vpeculiar_vector)
            self.translate(-distance_vector)
            self.rotate(axis_angle=("z", -self.ra))
            self.rotate(axis_angle=("y", self.dec))
        return

    def _init_pixcoords(self, datacube, origin=0):
        """
        Initialize pixel coordinates of the particles.

        Parameters
        ----------
        datacube : ~martini.datacube.DataCube
            The DataCube (including its WCS) for which to calculate coordinates.

        origin : int
            Index of the first pixel in the WCS (FITS-style is 1, python-style is 0).
        """
        skycoords_df_frame = self.skycoords.transform_to(datacube.coordinate_frame)
        spectralcoords_df_specsys = (
            self.spectralcoords.with_observer_stationary_relative_to(datacube.specsys)
        )
        # pixels indexed from 0 (not like in FITS!) for better use with numpy
        self.pixcoords = (
            np.vstack(
                datacube.wcs.sub(3).wcs_world2pix(
                    skycoords_df_frame.ra.to(datacube.units[0]),
                    skycoords_df_frame.dec.to(datacube.units[1]),
                    spectralcoords_df_specsys.to(datacube.units[2]),
                    origin,
                )
            )
            * U.pix
        )
        return

    def apply_mask(self, mask):
        """
        Remove particles from source arrays according to a mask.

        Parameters
        ----------
        mask : ~numpy.typing.ArrayLike
            Boolean mask. Remove particles with indices corresponding to
            ``False`` values from the source arrays.
        """

        if mask.size != self.npart:
            raise ValueError("Mask must have same length as particle arrays.")
        mask_sum = np.sum(mask)
        if mask_sum == 0:
            raise RuntimeError("No non-zero mHI source particles in target region.")
        self.npart = mask_sum
        if not self.T_g.isscalar:
            self.T_g = self.T_g[mask]
        if not self.mHI_g.isscalar:
            self.mHI_g = self.mHI_g[mask]
        self.coordinates_g = self.coordinates_g[mask]
        if self.skycoords is not None:
            self.skycoords = self.skycoords[mask]
        if self.spectralcoords is not None:
            self.spectralcoords = self.spectralcoords[mask]
        if self.pixcoords is not None:
            self.pixcoords = self.pixcoords[:, mask]
        if not self.hsm_g.isscalar:
            self.hsm_g = self.hsm_g[mask]
        return

    def rotate(self, axis_angle=None, rotmat=None, L_coords=None):
        """
        Rotate the source.

        The arguments correspond to different rotation types. Multiple types cannot be
        given in a single function call.

        Parameters
        ----------
        axis_angle : tuple
            First element one of {``"x"``, ``"y"``, ``"z"``} for the axis to rotate about,
            second element a :class:`~astropy.units.Quantity` with dimensions of angle,
            indicating the angle to rotate through (right-handed rotation).
        rotmat : ~numpy.typing.ArrayLike
            Rotation matrix with shape (3, 3).
        L_coords : tuple
            First element containing an inclination, second element an
            azimuthal angle (both :class:`~astropy.units.Quantity` instances with
            dimensions of angle). The routine will first attempt to identify
            a preferred plane based on the angular momenta of the central 1/3
            of particles in the source. This plane will then be rotated to lie
            in the 'y-z' plane, followed by a rotation by the azimuthal angle
            about its angular momentum pole (rotation about 'x'), and then
            inclined (rotation about 'y'). By default the position angle on the
            sky is 270 degrees, but if a third element is provided it sets the
            position angle (second rotation about 'x').
        """

        args_given = (axis_angle is not None, rotmat is not None, L_coords is not None)
        if np.sum(args_given) == 0:
            # no-op
            return
        elif np.sum(args_given) > 1:
            raise ValueError("Multiple rotations in a single call not allowed.")

        do_rot = np.eye(3)

        if axis_angle is not None:
            # rotation_matrix gives left-handed rotation, so transpose for right-handed
            do_rot = rotation_matrix(axis_angle[1], axis=axis_angle[0]).T.dot(do_rot)

        if rotmat is not None:
            do_rot = rotmat.dot(do_rot)

        if L_coords is not None:
            if len(L_coords) == 2:
                incl, az_rot = L_coords
                pa = 270 * U.deg
            elif len(L_coords) == 3:
                incl, az_rot, pa = L_coords
            do_rot = L_align(
                self.coordinates_g.get_xyz(),
                self.coordinates_g.differentials["s"].get_d_xyz(),
                self.mHI_g,
                frac=0.3,
                Laxis="x",
            ).dot(do_rot)
            # rotation_matrix gives left-handed rotation, so transpose for right-handed
            do_rot = rotation_matrix(az_rot, axis="x").T.dot(do_rot)
            do_rot = rotation_matrix(incl, axis="y").T.dot(do_rot)
            if incl >= 0:
                do_rot = rotation_matrix(pa - 90 * U.deg, axis="x").T.dot(do_rot)
            else:
                do_rot = rotation_matrix(pa - 270 * U.deg, axis="x").T.dot(do_rot)

        self.current_rotation = do_rot.dot(self.current_rotation)
        self.coordinates_g = self.coordinates_g.transform(do_rot)
        return

    def translate(self, translation_vector):
        """
        Translate the source.

        Note that the "line of sight" is along the 'x' axis.

        Parameters
        ----------
        translation_vector : ~astropy.units.Quantity
            :class:`~astropy.units.Quantity` with shape (3, ), with dimensions of length.
            Vector by which to offset the source particle coordinates.
        """

        self.coordinates_g = self.coordinates_g.translate(translation_vector)
        return

    def boost(self, boost_vector):
        """
        Apply an offset to the source velocity.

        Note that the "line of sight" is along the 'x' axis.

        Parameters
        ----------
        boost_vector : ~astropy.units.Quantity
            :class:`~astropy.units.Quantity` with shape (3, ), with dimensions of
            velocity.
            Vector by which to offset the source particle velocities.
        """

        self.coordinates_g.differentials["s"] = self.coordinates_g.differentials[
            "s"
        ].translate(boost_vector)
        return

    def save_current_rotation(self, fname):
        """
        Output current rotation matrix to file.

        This includes the rotations applied for RA and Dec. The rotation matrix can be
        applied to astropy coordinates (e.g. a
        :class:`~astropy.coordinates.representation.cartesian.CartesianRepresentation`) as
        ``coordinates.transform(np.loadtxt(fname))``.

        Parameters
        ----------
        fname : str
            File in which to save rotation matrix. A file handle can also be passed.
        """

        np.savetxt(fname, self.current_rotation)
        return

    def preview(
        self,
        max_points=5000,
        fig=1,
        lim=None,
        vlim=None,
        point_scaling="auto",
        title="",
        save=None,
    ):
        """
        Produce a figure showing the source particle coordinates and velocities.

        Makes a 3-panel figure showing the projection of the source as it will appear in
        the mock observation. The first panel shows the particles in the y-z plane,
        coloured by the x-component of velocity (MARTINI projects the source along the
        x-axis). The second and third panels are position-velocity diagrams showing the
        x-component of velocity against the y and z coordinates, respectively.

        Parameters
        ----------
        max_points : int, optional
            Maximum number of points to draw per panel, the particles will be randomly
            subsampled if the source has more. (Default: ``1000``)

        fig : int, optional
            Number of the figure in matplotlib, it will be created as ``plt.figure(fig)``.
            (Default: ``1``)

        lim : ~astropy.units.Quantity, optional
            :class:`~astropy.units.Quantity`, with dimensions of length.
            The coordinate axes extend from -lim to lim. If unspecified, the maximum
            absolute coordinate of particles in the source is used. (Default: ``None``)

        vlim : ~astropy.units.Quantity, optional
            :class:`~astropy.units.Quantity`, with dimensions of speed.
            The velocity axes and colour bar extend from ``-vlim`` to ``vlim``. If
            unspecified, the maximum absolute velocity of particles in the source is used.
            (Default: ``None``)

        point_scaling : str, optional
            By default points are scaled in size and transparency according to their HI
            mass and the smoothing length (loosely proportional to their surface
            densities, but with different scaling to achieve a visually useful plot). For
            some sources the automatic scaling may not give useful results, using
            ``point_scaling="fixed"`` will plot points of constant size without opacity.
            (Default: ``"auto"``)

        title : str, optional
            A title for the figure can be provided. (Default: ``""``)

        save : str, optional
            If provided, the figure is saved using ``plt.savefig(save)``. A ``.png`` or
            ``.pdf`` suffix is recommended. (Default: ``None``)

        Returns
        -------
        out : ~matplotlib.figure.Figure
            The preview figure.
        """
        import matplotlib.pyplot as plt

        # every Nth particle to plot at most max_points, or all particles
        lim = (
            max(
                np.max(np.abs(self.coordinates_g.y.to_value(U.kpc))),
                np.max(np.abs(self.coordinates_g.z.to_value(U.kpc))),
            )
            if lim is None
            else lim.to_value(U.kpc)
        )
        vlim = (
            np.max(
                np.abs(self.coordinates_g.differentials["s"].d_x.to_value(U.km / U.s))
            )
            if vlim is None
            else vlim.to_value(U.km / U.s)
        )
        cmask = np.logical_and.reduce(
            (
                np.abs(self.coordinates_g.y.to_value(U.kpc)) < lim,
                np.abs(self.coordinates_g.z.to_value(U.kpc)) < lim,
                np.abs(self.coordinates_g.differentials["s"].d_x.to_value(U.km / U.s))
                < vlim,
            )
        )
        nparts = cmask.sum()
        mask = np.arange(self.mHI_g.size)[cmask][:: max(nparts // max_points, 1)]
        # mass_factor = (
        #     1
        #     if (self.mHI_g.isscalar or mask.size <= 1)
        #     else (self.mHI_g[mask] / self.mHI_g[mask].max()).to_value(
        #         U.dimensionless_unscaled
        #     )
        # )
        hsm_factor = (
            1
            if (self.hsm_g.isscalar or mask.size <= 1)
            else (1 - (self.hsm_g[mask] / self.hsm_g[mask].max()) ** 0.1).to_value(
                U.dimensionless_unscaled
            )  # larger -> more transparent
            + (self.hsm_g[mask].min() == self.hsm_g[mask].max())
            * 0.5  # guard against getting all 0s
        )
        alpha = hsm_factor if point_scaling == "auto" else 1.0
        size_scale = (
            self.hsm_g.to_value(U.kpc) / lim
            if (self.hsm_g.isscalar or mask.size <= 1)
            else (self.hsm_g[mask].to_value(U.kpc) / lim)
        )
        size = 300 * size_scale if point_scaling == "auto" else 10
        print(f"{hsm_factor=} {size=}")
        fig = plt.figure(fig, figsize=(12, 4))
        fig.clf()
        fig.suptitle(title)

        # ----- MOMENT 1 -----
        sp1 = fig.add_subplot(1, 3, 1, aspect="equal")
        sp1.set_facecolor("#222222")
        scatter = sp1.scatter(
            self.coordinates_g.y[mask].to_value(U.kpc),
            self.coordinates_g.z[mask].to_value(U.kpc),
            c=self.coordinates_g.differentials["s"].d_x[mask].to_value(U.km / U.s),
            marker="o",
            cmap="coolwarm",
            edgecolor="None",
            s=size,
            alpha=alpha,
            vmin=-vlim,
            vmax=vlim,
            zorder=0,
        )
        sp1.plot([0], [0], marker="+", ls="None", mec="grey", ms=6, zorder=1)
        sp1.set_xlabel(r"$y\,[\mathrm{kpc}]$")
        sp1.set_ylabel(r"$z\,[\mathrm{kpc}]$")
        sp1.set_xlim((lim, -lim))
        sp1.set_ylim((-lim, lim))
        cb = fig.colorbar(mappable=scatter, ax=sp1, orientation="horizontal")
        cb.set_label(r"$v_x\,[\mathrm{km}\,\mathrm{s}^{-1}]$")

        # ----- PV Y -----
        sp2 = fig.add_subplot(1, 3, 2)
        sp2.set_facecolor("#222222")
        sp2.scatter(
            self.coordinates_g.y[mask].to_value(U.kpc),
            self.coordinates_g.differentials["s"].d_x[mask].to_value(U.km / U.s),
            c="white",
            edgecolors="None",
            marker="o",
            s=size,
            alpha=alpha,
            zorder=0,
        )
        sp2.plot([0], [0], marker="+", ls="None", mec="red", ms=6, zorder=1)
        sp2.set_xlim((lim, -lim))
        sp2.set_ylim((-vlim, vlim))
        sp2.set_xlabel(r"$y\,[\mathrm{kpc}]$")
        sp2.set_ylabel(r"$v_x\,[\mathrm{km}\,\mathrm{s}^{-1}]$")

        # ----- PV Z -----
        sp3 = fig.add_subplot(1, 3, 3)
        sp3.set_facecolor("#222222")
        sp3.scatter(
            self.coordinates_g.z[mask].to_value(U.kpc),
            self.coordinates_g.differentials["s"].d_x[mask].to_value(U.km / U.s),
            c="white",
            edgecolors="None",
            marker="o",
            s=size,
            alpha=alpha,
            zorder=0,
        )
        sp3.plot([0], [0], marker="+", ls="None", mec="red", ms=6, zorder=1)
        sp3.set_xlim((-lim, lim))
        sp3.set_ylim((-vlim, vlim))
        sp3.set_xlabel(r"$z\,[\mathrm{kpc}]$")
        sp3.set_ylabel(r"$v_x\,[\mathrm{km}\,\mathrm{s}^{-1}]$")

        fig.subplots_adjust(wspace=0.3)
        if save is not None:
            plt.savefig(save)
        return fig



import pytest
import os
import numpy as np
from astropy import wcs
from astropy import units as U
from martini import DataCube
from martini.datacube import HIfreq


def check_wcs_match(wcs1, wcs2):
    assert set(wcs1.to_header().keys()) == set(wcs2.to_header().keys())
    for k, v, _ in wcs1.to_header().cards:
        assert v == wcs2.to_header()[k]


class TestDeprecationWarnings:
    def test_deprecation_warnings(self):
        """
        Check that we get deprecation warnings where we expect them.
        """
        with pytest.warns(DeprecationWarning):
            dc = DataCube(velocity_centre=0 * U.km / U.s)
        with pytest.warns(DeprecationWarning):
            dc.velocity_channels()
        with pytest.warns(DeprecationWarning):
            dc.freq_channels()


class TestDataCube:
    def test_invalid_channel_units(self):
        """
        Check that we get an error if the channel_width doesn't have valid units.
        """
        with pytest.raises(ValueError, match="Channel width must have"):
            DataCube(channel_width=1 * U.K)

    def test_invalid_specsys(self):
        """
        Check that we get an error for an invalid specsys.
        """
        with pytest.raises(ValueError, match="Supported specsys values are"):
            DataCube(specsys="rubbish")

    def test_datacube_dimensions(self):
        """
        Check that dimensions are as requested.
        """
        datacube = DataCube(n_px_x=10, n_px_y=11, n_channels=12)
        expected_shape = (10, 11, 12, 1) if datacube.stokes_axis else (10, 11, 12)
        assert datacube._array.shape == expected_shape

    def test_channel_mids(self, dc_zeros):
        """
        Check that first and last channel mids are spaced as expected.
        """
        bandwidth = np.abs(dc_zeros.channel_mids[-1] - dc_zeros.channel_mids[0])
        assert U.isclose(bandwidth, (dc_zeros.n_channels - 1) * dc_zeros.channel_width)

    def test_channel_edges(self, dc_zeros):
        """
        Check that first and last channel edges are spaced as expected.
        """
        bandwidth = np.abs(dc_zeros.channel_edges[-1] - dc_zeros.channel_edges[0])
        assert bandwidth == dc_zeros.n_channels * dc_zeros.channel_width

    def test_iterators(self, dc_zeros):
        """
        Check that iterators over slices give us expected lengths.
        """
        assert len(list(dc_zeros.channel_maps)) == dc_zeros.n_channels
        assert len(list(dc_zeros.spatial_slices)) == dc_zeros.n_channels
        assert len(list(dc_zeros.spectra)) == dc_zeros.n_px_x * dc_zeros.n_px_y

    def test_freq_channels(self, dc_zeros):
        """
        Check that frequency channels match WCS.
        """
        spec_unit = U.Unit(dc_zeros.wcs.wcs.cunit[dc_zeros.wcs.wcs.spec], format="fits")
        mids = (
            dc_zeros.wcs.sub(("spectral",)).all_pix2world(
                np.arange(dc_zeros.n_channels), 0
            )
        ) * spec_unit
        edges = (
            dc_zeros.wcs.sub(("spectral",)).all_pix2world(
                np.arange(dc_zeros.n_channels + 1) - 0.5, 0
            )
        ) * spec_unit
        assert U.allclose(
            mids.to(U.Hz, equivalencies=U.doppler_radio(HIfreq)),
            dc_zeros.frequency_channel_mids,
        )
        assert U.allclose(
            edges.to(U.Hz, equivalencies=U.doppler_radio(HIfreq)),
            dc_zeros.frequency_channel_edges,
        )

    def test_velocity_channels(self, dc_zeros):
        """
        Check that velocity channels match WCS.
        """
        spec_unit = U.Unit(dc_zeros.wcs.wcs.cunit[dc_zeros.wcs.wcs.spec], format="fits")
        mids = (
            dc_zeros.wcs.sub(("spectral",)).all_pix2world(
                np.arange(dc_zeros.n_channels), 0
            )
        ) * spec_unit
        edges = (
            dc_zeros.wcs.sub(("spectral",)).all_pix2world(
                np.arange(dc_zeros.n_channels + 1) - 0.5, 0
            )
        ) * spec_unit
        assert U.allclose(
            mids.to(U.m / U.s, equivalencies=U.doppler_radio(HIfreq)),
            dc_zeros.velocity_channel_mids,
        )
        assert U.allclose(
            edges.to(U.m / U.s, equivalencies=U.doppler_radio(HIfreq)),
            dc_zeros.velocity_channel_edges,
        )

    def test_add_pad(self, dc_zeros):
        """
        Check that adding pad gives desired shape.
        """
        old_shape = dc_zeros._array.shape
        pad = (2, 3)
        dc_zeros.add_pad(pad)
        expected_shape = (
            old_shape[0] + 2 * pad[0],
            old_shape[1] + 2 * pad[1],
            old_shape[2],
        )
        if dc_zeros.stokes_axis:
            expected_shape = expected_shape + (old_shape[3],)
        assert dc_zeros._array.shape == expected_shape
        assert dc_zeros.padx == pad[0]
        assert dc_zeros.pady == pad[1]

    def test_add_pad_already_padded(self, dc_zeros):
        """
        Check that we can't double-pad.
        """
        pad = (2, 3)
        dc_zeros.add_pad(pad)
        with pytest.raises(RuntimeError, match="Tried to add padding"):
            dc_zeros.add_pad(pad)

    def test_drop_pad(self, dc_zeros):
        """
        Check that we get expected shape when dropping padding.
        """
        initial_shape = dc_zeros._array.shape
        pad = (2, 3)
        dc_zeros.add_pad(pad)
        old_shape = dc_zeros._array.shape
        dc_zeros.drop_pad()
        expected_shape = (
            old_shape[0] - 2 * pad[0],
            old_shape[1] - 2 * pad[1],
            old_shape[2],
        )
        if dc_zeros.stokes_axis:
            expected_shape = expected_shape + (old_shape[3],)
        assert dc_zeros._array.shape == initial_shape
        assert dc_zeros._array.shape == expected_shape
        assert dc_zeros.padx == 0
        assert dc_zeros.pady == 0

    def test_drop_pad_already_dropped(self, dc_zeros):
        """
        Check that dropping already dropped pad gives no change.
        """
        assert dc_zeros.padx == 0
        assert dc_zeros.pady == 0
        dc_zeros.drop_pad()
        assert dc_zeros.padx == 0
        assert dc_zeros.pady == 0

    @pytest.mark.parametrize("with_pad", (False, True))
    def test_copy(self, dc_random, with_pad):
        """
        Check that copying a datacube copies all required information.
        """
        if with_pad:
            dc_random.add_pad((3, 3))
        copy = dc_random.copy()
        for attr in (
            "n_px_x",
            "n_px_y",
            "n_channels",
            "padx",
            "pady",
        ):
            assert getattr(dc_random, attr) == getattr(copy, attr)
        for attr in (
            "px_size",
            "channel_width",
            "spectral_centre",
            "ra",
            "dec",
        ):
            assert U.isclose(
                getattr(dc_random, attr),
                getattr(copy, attr),
                atol=1e-6 * getattr(dc_random, attr).unit,
            )
        for attr in (
            "_channel_edges",
            "_channel_mids",
            "_array",
        ):
            if getattr(dc_random, attr) is not None:
                assert U.allclose(getattr(dc_random, attr), getattr(copy, attr))
            else:
                assert getattr(copy, attr) is None
        check_wcs_match(dc_random.wcs, copy.wcs)

    @pytest.mark.parametrize("with_pad", (False, True))
    def test_save_and_load_state(self, dc_random, with_pad):
        """
        Check that we can recover a datacube from a save file.
        """
        pytest.importorskip("h5py", reason="h5py (optional dependency) not available.")
        try:
            if with_pad:
                dc_random.add_pad((3, 3))
            dc_random.save_state("test_savestate.hdf5", overwrite=True)
            loaded = DataCube.load_state("test_savestate.hdf5")
            for attr in (
                "n_px_x",
                "n_px_y",
                "n_channels",
                "padx",
                "pady",
            ):
                assert getattr(dc_random, attr) == getattr(loaded, attr)
            for attr in (
                "px_size",
                "channel_width",
                "spectral_centre",
                "ra",
                "dec",
            ):
                assert U.isclose(
                    getattr(dc_random, attr),
                    getattr(loaded, attr),
                    atol=1e-6 * getattr(dc_random, attr).unit,
                )
            for attr in (
                "channel_edges",
                "channel_mids",
                "_array",
            ):
                assert U.allclose(getattr(dc_random, attr), getattr(loaded, attr))
            check_wcs_match(dc_random.wcs, loaded.wcs)
        except Exception:
            raise
        finally:
            os.remove("test_savestate.hdf5")

    def test_init_with_mixed_spectral_centre_and_channel_width_units(self):
        """
        Check that we can specify channel spacing and central channel in mixed units.
        """
        const_kwargs = dict(
            n_px_x=16,
            n_px_y=16,
            n_channels=16,
        )
        spectral_centre = 3 * 70 * U.km / U.s
        channel_width = 4 * U.km / U.s
        f_channel_width = np.abs(
            (spectral_centre + 0.5 * channel_width).to(
                U.Hz, equivalencies=U.doppler_radio(HIfreq)
            )
            - (spectral_centre - 0.5 * channel_width).to(
                U.Hz, equivalencies=U.doppler_radio(HIfreq)
            )
        )
        f_spectral_centre = spectral_centre.to(
            U.Hz, equivalencies=U.doppler_radio(HIfreq)
        )
        dc_vv = DataCube(
            spectral_centre=spectral_centre,
            channel_width=channel_width,
            **const_kwargs,
        )
        dc_vf = DataCube(
            spectral_centre=spectral_centre,
            channel_width=f_channel_width,
            **const_kwargs,
        )
        dc_fv = DataCube(
            spectral_centre=f_spectral_centre,
            channel_width=channel_width,
            **const_kwargs,
        )
        dc_ff = DataCube(
            spectral_centre=f_spectral_centre,
            channel_width=f_channel_width,
            **const_kwargs,
        )
        # expect channels to match where units of channel_width match
        # channel width as velocity:
        assert U.allclose(
            dc_vv.channel_mids,
            dc_fv.channel_mids,
        )
        # channel width as frequency:
        assert U.allclose(
            dc_ff.channel_mids,
            dc_vf.channel_mids,
        )

    def test_channel_spacing(self, dc_zeros):
        """
        Expect channels to be equally spaced in units matching channel_width, check that
        this is the case.
        """
        assert U.get_physical_type(dc_zeros.channel_width) == U.get_physical_type(
            dc_zeros.channel_mids
        )
        assert U.get_physical_type(dc_zeros.channel_width) == U.get_physical_type(
            dc_zeros.channel_edges
        )
        assert U.allclose(
            np.diff(np.diff(dc_zeros.channel_edges)),
            0 * dc_zeros.channel_width.unit,
            atol=1e-5 * dc_zeros.channel_width.unit,
        )
        assert U.allclose(
            np.diff(np.diff(dc_zeros.channel_edges)),
            0 * dc_zeros.channel_width.unit,
            atol=1e-5 * dc_zeros.channel_width.unit,
        )
        if U.get_physical_type(dc_zeros.channel_width) == "frequency":
            assert U.allclose(
                np.diff(np.diff(dc_zeros.frequency_channel_edges)),
                0 * U.Hz,
                atol=1e-5 * U.Hz,
            )
            assert U.allclose(
                np.diff(np.diff(dc_zeros.frequency_channel_mids)),
                0 * U.Hz,
                atol=1e-5 * U.Hz,
            )
        elif U.get_physical_type(dc_zeros.channel_width) == "velocity":
            assert U.allclose(
                np.diff(np.diff(dc_zeros.velocity_channel_edges)),
                0 * U.m / U.s,
                atol=1e-5 * U.m / U.s,
            )
            assert U.allclose(
                np.diff(np.diff(dc_zeros.velocity_channel_mids)),
                0 * U.m / U.s,
                atol=1e-5 * U.m / U.s,
            )


class TestDataCubeFromWCS:

    def test_specsys_case_insensitive(self, dc_random):
        """
        FITS specsys usually uppercase but astropy has equivalents in lowercase.
        Check that we convert successfully.
        """
        dc_random.wcs.wcs.specsys = "ICRS"
        from_wcs = DataCube.from_wcs(dc_random.wcs)
        assert from_wcs.specsys == "icrs"

    def test_unrecognized_specsys(self, dc_random):
        """
        Check that we get an error for a specsys that we can't understand.
        """
        dc_random.wcs.wcs.specsys = "rubbish"
        with pytest.raises(ValueError, match="not yet supported by MARTINI"):
            DataCube.from_wcs(dc_random.wcs)

    def test_invalid_specsys(self, dc_random):
        """
        Check that we get an error for an invalid specsys.
        """
        with pytest.raises(ValueError, match="Supported specsys values are"):
            DataCube.from_wcs(dc_random.wcs, specsys="rubbish")

    def test_rectangular_pixels_rejected(self, dc_random):
        """
        Check that we refuse to accept rectangular pixels.
        """
        if dc_random.stokes_axis:
            dc_random.wcs.wcs.cdelt = (
                dc_random.wcs.wcs.cdelt[0],
                dc_random.wcs.wcs.cdelt[1] + 1,
                dc_random.wcs.wcs.cdelt[2],
                dc_random.wcs.wcs.cdelt[3],
            )
        else:
            dc_random.wcs.wcs.cdelt = (
                dc_random.wcs.wcs.cdelt[0],
                dc_random.wcs.wcs.cdelt[1] + 1,
                dc_random.wcs.wcs.cdelt[2],
            )
        with pytest.raises(ValueError, match="RA and Dec axes do not match"):
            DataCube.from_wcs(dc_random.wcs)

    def test_restfrq_missing(self, dc_random):
        """
        Check that we warn if the WCS doesn't specify a RESTFRQ.
        """
        dc_random.wcs.wcs.restfrq = 0.0
        with pytest.warns(UserWarning, match="Input WCS did not specify RESTFRQ"):
            DataCube.from_wcs(dc_random.wcs)

    def test_consistent_with_direct(self, dc_random):
        """
        Check that extracting WCS from a constructed DataCube and constructing
        a DataCube from that WCS is consistent with the original DataCube.

        Note that we shouldn't expect this to reproduce a padded cube, or use
        a WCS from a padded cube!
        """
        from_wcs = DataCube.from_wcs(dc_random.wcs)
        for attr in (
            "n_px_x",
            "n_px_y",
            "n_channels",
            "padx",
            "pady",
        ):
            assert getattr(dc_random, attr) == getattr(from_wcs, attr)
        for attr in (
            "px_size",
            "channel_width",
            "spectral_centre",
            "ra",
            "dec",
        ):
            assert U.isclose(
                getattr(dc_random, attr),
                getattr(from_wcs, attr),
                atol=1e-6 * getattr(dc_random, attr).unit,
            )
        # don't test for arrays matching, they should not:
        for attr in (
            "channel_edges",
            "channel_mids",
        ):
            assert U.allclose(
                getattr(dc_random, attr),
                getattr(from_wcs, attr),
                atol=1e-6 * getattr(dc_random, attr).unit,
            )
        check_wcs_match(dc_random.wcs, from_wcs.wcs)

    @pytest.mark.parametrize(
        "sample_header",
        (
            "IC_2574_NA_CUBE_THINGS.HEADER",
            "comb_10tracks_J1337_28_HI_r10_t90_mg095_2.image.header",
            "WALLABY_PILOT_CUTOUT_APPROX.HEADER",
            # as previous but with axes reordered:
            "comb_10tracks_J1337_28_HI_r10_t90_mg095_2.image.header.permuted",
        ),
    )
    def test_sample_headers(self, sample_header):
        """
        Check that some real-world FITS headers can be used to initialize a DataCube.
        """
        with open(os.path.join("tests/data/", sample_header), "r") as f:
            hdr = f.read()
        with pytest.warns(wcs.FITSFixedWarning):
            hdr_wcs = wcs.WCS(hdr)
        if hdr_wcs.wcs.specsys == "":
            with pytest.warns(
                UserWarning,
                match="Input WCS did not specify 'SPECSYS'",
            ):
                dc = DataCube.from_wcs(hdr_wcs)
        elif hdr_wcs.wcs.specsys == "BARYCENT":
            with pytest.warns(
                UserWarning, match="Assuming ICRS barycentric reference system."
            ):
                dc = DataCube.from_wcs(hdr_wcs)
        else:
            dc = DataCube.from_wcs(hdr_wcs)
        assert dc.stokes_axis == (hdr_wcs.naxis == 4)
        centre_coords = hdr_wcs.all_pix2world(
            [[n_px // 2 + (1 + n_px % 2) / 2 for n_px in hdr_wcs.pixel_shape]],
            1,  # origin, i.e. index pixels from 1
        ).squeeze()
        for i, (centre_coord, unit, spacing, len_ax) in enumerate(
            zip(
                centre_coords,
                hdr_wcs.world_axis_units,
                hdr_wcs.wcs.cdelt,
                hdr_wcs.pixel_shape,
            )
        ):
            if i == hdr_wcs.wcs.lng:
                assert U.isclose(dc.px_size, -spacing * U.Unit(unit, format="fits"))
                assert dc.n_px_x == len_ax
                assert U.isclose(dc.ra, centre_coord * U.Unit(unit, format="fits"))
            elif i == hdr_wcs.wcs.lat:
                assert U.isclose(dc.px_size, spacing * U.Unit(unit, format="fits"))
                assert dc.n_px_y == len_ax
                assert U.isclose(dc.dec, centre_coord * U.Unit(unit, format="fits"))
            elif i == hdr_wcs.wcs.spec:
                # This breaks if spacing is in Hz,
                # need the frequency difference at the velocity centre:
                hdr_specref = centre_coord * U.Unit(unit, format="fits")
                assert U.isclose(
                    dc.channel_width,
                    np.abs(
                        (hdr_specref + spacing * U.Unit(unit, format="fits")).to(
                            dc.channel_width.unit, equivalencies=U.doppler_radio(HIfreq)
                        )
                        - (hdr_specref).to(
                            dc.channel_width.unit, equivalencies=U.doppler_radio(HIfreq)
                        )
                    ),
                )
                assert dc.n_channels == len_ax
                assert U.isclose(
                    dc.spectral_centre,
                    hdr_specref.to(
                        dc.spectral_centre.unit, equivalencies=U.doppler_radio(HIfreq)
                    ),
                )



import typing as T
import astropy.units as U
import abc
from _typeshed import Incomplete
from abc import abstractmethod
from martini.datacube import DataCube as DataCube
from martini.beams import _BaseBeam
from numpy.random._generator import Generator

class _BaseNoise(metaclass=abc.ABCMeta):
    __metaclass__: Incomplete
    seed: int
    rng: Generator

    def __init__(self, seed: T.Optional[int] = ...) -> None: ...
    @abstractmethod
    def generate(self, datacube: DataCube, beam: _BaseBeam): ...
    def reset_rng(self) -> None: ...

class GaussianNoise(_BaseNoise):
    rms: U.Quantity[U.Jy * U.beam**-1]

    def __init__(
        self, rms: U.Quantity[U.Jy * U.beam**-1] = ..., seed: T.Optional[int] = ...
    ) -> None: ...
    def generate(
        self, datacube: DataCube, beam: _BaseBeam
    ) -> U.Quantity[U.Jy * U.arcsec**-2]: ...



import typing as T
from _typeshed import Incomplete
import astropy.units as U
import abc
import numpy as np
from abc import abstractmethod
from martini.datacube import DataCube as DataCube

class _BaseBeam(metaclass=abc.ABCMeta):
    __metaclass__: Incomplete
    bmaj: U.Quantity[U.arcsec]
    bmin: U.Quantity[U.arcsec]
    bpa: U.Quantity[U.deg]
    px_size: T.Optional[U.Quantity[U.arcsec]]
    kernel: T.Optional[U.Quantity[U.dimensionless_unscaled]]
    area: U.Quantity[U.arcsec**2]

    def __init__(
        self,
        bmaj: U.Quantity[U.arcsec] = ...,
        bmin: U.Quantity[U.arcsec] = ...,
        bpa: U.Quantity[U.deg] = ...,
    ) -> None: ...
    def needs_pad(self) -> T.Tuple[int, int]: ...

    vel: U.Quantity[U.km / U.s]
    ra: U.Quantity[U.deg]
    dec: U.Quantity[U.deg]

    def init_kernel(self, datacube: DataCube) -> None: ...
    @abstractmethod
    def f_kernel(self): ...
    @abstractmethod
    def kernel_size_px(self): ...
    @abstractmethod
    def init_beam_header(self): ...

class GaussianBeam(_BaseBeam, metaclass=abc.ABCMeta):
    truncate: float

    def __init__(
        self,
        bmaj: U.Quantity[U.arcsec] = ...,
        bmin: U.Quantity[U.arcsec] = ...,
        bpa: U.Quantity[U.deg] = ...,
        truncate: float = ...,
    ) -> None: ...
    def f_kernel(
        self,
    ) -> T.Callable[
        [T.Union[float, np.ndarray], T.Union[float, np.ndarray]], U.Quantity
    ]: ...
    def kernel_size_px(self) -> T.Tuple[int, int]: ...



import astropy.units as U
from astropy.coordinates import CartesianRepresentation, CartesianDifferential

def translate(
    cls: CartesianRepresentation, translation_vector: U.Quantity[U.kpc]
) -> CartesianRepresentation: ...
def translate_d(
    cls: CartesianDifferential, translation_vector: U.Quantity[U.km / U.s]
) -> CartesianDifferential: ...
